#include "game.hpp"
#include <algorithm>
#include <chrono>
#include <cstdlib>
#include <fstream>
#include <ios>
#include <iostream>
#include <mutex>
#include <thread>

GameState::GameState() : word(), tries(0), guesses() {}

GameState::GameState(const std::string &w) : word(w), tries(0), guesses() {}

void GameState::setWord(const std::string &w) { word = w; }

std::array<GameState::matchType, WORD_LENGTH>
GameState::guess(const std::string &guessedWord) {
  std::array<matchType, WORD_LENGTH> matchStatus;
  // iterate over the word character by character
  for (std::string::size_type i = 0; i < guessedWord.size(); i++) {
    if (word[i] == guessedWord[i]) {
      matchStatus[i] = EXACT_MATCH;
    } else if (word.find(guessedWord[i])) {
      matchStatus[i] = MATCH;
    } else {
      matchStatus[i] = NO_MATCH;
    }
  }
  return matchStatus;
}

std::string GameState::serialize(bool newWord) {
  auto gs = new wordle_de::GameState();
  gs->set_newword(newWord);
  gs->set_tries(tries);
  auto msg = wordle_de::Message();
  msg.set_allocated_gamestate(gs);
  return msg.SerializeAsString();
}

void Engine::createInstance() {
  if (instance == nullptr) {
    instance = new Engine();
  }
}

void Engine::createInstance(const std::string &fileName) {
  if (instance == nullptr) {
    instance = new Engine(fileName);
  }
}

Engine *Engine::getInstance() { return instance; }

void Engine::destroyInstance() { delete instance; }

void Engine::setNewWord() {
  if (!currentWord.empty()) {
    pastWords.push_back(currentWord);
  }
  // read random line from file
  std::ifstream file(wordFile);
  if (file.is_open()) {
    // find number of lines in file
    file.unsetf(std::ios_base::skipws);
    auto lines = std::count(std::istream_iterator<char>(file),
                            std::istream_iterator<char>(), '\n');
    file.setf(std::ios_base::skipws);
    file.clear();
    file.seekg(0, std::ios::beg);
    auto randLine = getRandomLine(lines);
    for (int i = 0; i < randLine - 1; i++)
      file.ignore(std::numeric_limits<std::streamsize>::max(), '\n');
    std::string word;
    file >> word;
    file.close();
    setWord(word);
  }
  std::cout << "New Word: " << currentWord << std::endl;
  std::cout << "Old words:" << std::endl;
  for (const auto &w : pastWords) {
    std::cout << w << std::endl;
  }
}

void Engine::setWord(const std::string &w) {
  std::lock_guard<std::mutex> lock(mut);
  currentWord = w;
}
std::string Engine::getWord() {
  std::lock_guard<std::mutex> lock(mut);
  return currentWord;
}

void Engine::start(unsigned int interval) {
  if (!started) {
    std::thread([interval, this]() {
      while (true) {
        setNewWord();
        for (auto cb : callbacks) {
          cb(currentWord);
        }
        std::this_thread::sleep_for(std::chrono::seconds(interval));
      }
    }).detach();
    started = true;
  }
}

unsigned int Engine::getRandomLine(unsigned int max) {
  srand(time(NULL));
  return rand() % max + 1;
}

void Engine::registerNewWordCallback(
    std::function<void(const std::string &)> cb) {
  std::lock_guard<std::mutex> lock(mut);
  callbacks.push_back(cb);
}

Engine *Engine::instance;
